 /*
  * UADE
  *
  * Support for AmigaOS/MorphOS
  * 
  * Copyright 2003 Harry Sintonen
  */

#include <uade-os-amiga.h>
#include <uadeconfig.h>
#include <sys/types.h>
#include <sys/time.h>
#include <dirent.h>
#include <ctype.h>

#include <errno.h>
#include <dos/dos.h>
#include <devices/timer.h>
#include <workbench/startup.h>

#include <proto/exec.h>
#include <proto/dos.h>

#include "strlrep.c"

static int os_debug = 0;

static int unix_soundfd;
static int unix_usingoutpipe = 0;
void (*sig2handler)(int) = NULL;

void uade_create_signalhandler(void (*sighandler)(int), int sigtype) {
  if (sigtype == 2)
  {
    sig2handler = sighandler;
    return;
  }
  errno = ENOSYS;
}


void uade_create_var_pid(void) {
  errno = ENOSYS;
}

void uade_exit(int status) {
  switch (status) {
    case -1:
    case  1:
      status = RETURN_ERROR; break;
  }

  exit(status);
}

/* first element of args is the executable file */
/* last element of args table has to be a NULL pointer */
pid_t uade_fork_exec(char **args) {
  errno = ENOSYS;
  return -1;
}


extern struct WBStartup *_WBenchMsg;

static int ami_getprogdir(char *buf, int maxsize)
{
  if (_WBenchMsg) {
    if (NameFromLock(_WBenchMsg->sm_ArgList[0].wa_Lock, buf, maxsize)) {
      return 1;
    }
  }
  else {
    BPTR lock = GetProgramDir();
    if (lock && NameFromLock(lock, buf, maxsize)) {
      return 1;
    }
  }

  return 0;
}

FILE *uade_get_file(int item) {
  char post[1024];
  FILE *testfile;
  post[0] = 0;
  if(!ami_getprogdir(post, sizeof(post))) {
    fprintf(stderr, "uade: warning: can't get PROGDIR. using '' instead\n");
    post[0] = 0;
  }
  switch(item) {
  case UADE_PATH_SCORE:
    if (!AddPart(post, "score", sizeof(post)))
      post[0] = 0;
    break;
  case UADE_PATH_UAERC:
    if (!AddPart(post, "uaerc", sizeof(post)))
      post[0] = 0;
    break;
  case UADE_PATH_UADE:
    if (!AddPart(post, "uade", sizeof(post)))
      post[0] = 0;
    break;
  case UADE_PATH_FORMATSFILE:
    if (!AddPart(post, "players/uadeformats", sizeof(post)))
      post[0] = 0;
    break;
  default:
    fprintf(stderr, "uade: failed to get path of %d\n", item);
    return 0;
  }
  if(post[0]) {
    testfile = fopen(post, "r");
    if(testfile) {
      if(os_debug)
	fprintf(stderr, "uade: get file: success: path = %s\n", post);
      return testfile;
    }
  }
  fprintf(stderr, "uade: failed to get path of %s\n", post);
  return 0;
}


int uade_get_path(char *path, int item, int maxlen) {
  char post[1024];
  FILE *testfile;
  DIR *testdir;
  post[0] = 0;
  path[0] = 0;
  if(!ami_getprogdir(post, sizeof(post))) {
    fprintf(stderr, "uade: warning: can't get PROGDIR. using '' instead\n");
    post[0] = 0;
  }
  switch(item) {
  case UADE_PATH_SCORE:
    if (!AddPart(post, "score", sizeof(post)))
      post[0] = 0;
    break;
  case UADE_PATH_UAERC:
    if (!AddPart(post, "uaerc", sizeof(post)))
      post[0] = 0;
    break;
  case UADE_PATH_UADE:
    if (!AddPart(post, "uade", sizeof(post)))
      post[0] = 0;
    break;
  case UADE_PATH_PLAYERDIR:
    if (!AddPart(post, "players/", sizeof(post))) {
      post[0] = 0;
      break;
    }
    testdir = opendir(post);
    if(testdir) {
      closedir(testdir);
      strlcpy(path, post, maxlen);
      if (AddPart(post, "uadeformats", sizeof(post))) {
	testfile = fopen(post, "r");
	if(testfile) {
	  fclose(testfile); testfile = 0;
	  if(os_debug)
	    fprintf(stderr, "success: path = %s\n", path);
	  return 1;
	}
      }
    }
    fprintf(stderr, "uade: couldn't get uade playerdir path\n");
    return 0;
  case UADE_PATH_FORMATSFILE:
    if (!AddPart(post, "players/uadeformats", sizeof(post)))
      post[0] = 0;
    break;
  default:
    fprintf(stderr, "uade: failed to get path of %d\n", item);
    return 0;
  }
  if(post[0]) {
    testfile = fopen(post, "r");
    if(testfile) {
      fclose(testfile); testfile = 0;
      strlcpy(path, post, maxlen);
      if(os_debug)
	fprintf(stderr, "uade: success: path = %s\n", path);
      return 1;
    }
  }
  fprintf(stderr,"uade: failed to get path of %s\n", post);
  return 0;
}


int uade_get_postfix(char *dst, char *filename, int maxlen) {
  char *baseptr;
  if (maxlen == 0)
    return 0;
  baseptr = strrchr(filename, '/');
  if (baseptr) {
    baseptr++;
  } else {
    baseptr = filename;
  }
  baseptr = strrchr(baseptr, (int) '.');
  if (!baseptr) {
    return 0;
  }
  baseptr++;
  if (((int) strlen(baseptr)) >= maxlen) {
    return 0;
  }
  strcpy(dst, baseptr); /* will not overflow */
  return 1;
}


int uade_get_prefix(char *dst, char *filename, int maxlen) {
  int i, ret;
  char *baseptr;
  if (maxlen == 0)
    return 0;
  baseptr = strrchr(filename, '/');
  if (baseptr) {
    baseptr++;
  } else {
    baseptr = filename;
  }
  if (!strchr(baseptr, (int) '.')) {
    return 0;
  }
  ret = 0;
  for (i=0;;i++) {
    if (i >= maxlen)
      break;
    if (baseptr[i] == '.') {
      ret = 1;
      break;
    }
    dst[i] = baseptr[i];
  }
  dst[i] = 0;
  return ret;
}


int uade_get_temp_name(char *tempname, int maxlen) {
  errno = ENOSYS;
  return -1;
}


int uade_is_regular_file(const char *filename) {
  UBYTE _fib[sizeof(struct FileInfoBlock) + 3];
  struct FileInfoBlock *fib = (void *) (((ULONG) _fib + 3) & ~3);
  BPTR lock;
  int ret = 0;

  lock = Lock(filename, ACCESS_READ);
  if (lock) {
    if (Examine(lock, fib)) {
      if (fib->fib_DirEntryType < 0) {
        ret = 1;
      }
    }
    UnLock(lock);
  }

  return ret;
}


FILE *uade_open_amiga_file(char *amigafilename) {
  return fopen(amigafilename, "rb");
}


void uade_init_outpipe(char *parameter) {
  unix_usingoutpipe = 1;
  unix_soundfd = atoi(parameter);
}


int uade_send_signal(pid_t pid, int sigtype) {
  errno = ENOSYS;
  return -1;
}

#ifdef __libnix__

/* posix emulation - VERY limited edition */

static int atexittemp = 0;

static void mkstemp_clean(void) {
  char fake[] = "/tmp/fooXXXXXX";
  int fd = mkstemp(fake);
  if (fd > -1) {
    close(fd);
    remove(fake);
  }
}

int mkstemp(char *template) {
  int fd = - 1;
  if (template) {
    int len = strlen(template);
    if (len >= 12 && memcmp(template + len - 6, "XXXXXX", 6) == 0) {
      /* be silly and always generate the same file. this ONLY makes
         sense here because libnix mkstemp doens't delete the file
         at close. THIS IS NOT GENERIC SOLUTION! - Piru */
      int id = (int) FindTask(NULL) & 0xffffff;
      sprintf(template, "T:uade%06x", id);
      fd = open(template, O_TRUNC | O_RDWR | O_CREAT | O_EXCL);
      if (fd && !atexittemp) {
        atexittemp = 1;
        atexit(mkstemp_clean);
      }
    }
    else {
      errno = EINVAL;
    }
  }
  return fd;
}

FILE *popen(const char *command, const char *type) {

#warning FIXME: Use fopen with APIPE: ?

  errno = ENOSYS;
  return NULL;
}

int pclose(FILE *stream) {
  errno = ENOSYS;
  return -1;
}

#endif /* __libnix__ */


static int ami_timerinit = 0;
static struct timerequest *ami_tr;
static int ami_timerdev = -1;

static void ami_timercleanup(void) {
  if (ami_timerdev == 0) {
    ami_tr->tr_node.io_Message.mn_ReplyPort = NULL;
    CloseDevice(&ami_tr->tr_node);
    ami_timerdev = -1;
  }
  if (ami_tr) {
    DeleteIORequest(&ami_tr->tr_node);
    ami_tr = NULL;
  }
}

static void ami_sleep(ULONG unit, struct timeval *tv) {
  struct MsgPort replyport;
  ULONG sigs;

  replyport.mp_Node.ln_Type = NT_MSGPORT;
  replyport.mp_Flags        = PA_SIGNAL;
  replyport.mp_SigBit       = SIGB_SINGLE;
  replyport.mp_SigTask      = FindTask(NULL);
  NewList(&replyport.mp_MsgList);

  if (!ami_timerinit) {
    ami_timerinit = 1;

    do {
      if (atexit(ami_timercleanup)) {
        ami_tr = CreateIORequest(&replyport, sizeof(*ami_tr));
        if (ami_tr) {
          ami_timerdev = OpenDevice(TIMERNAME, unit, &ami_tr->tr_node, 0);
          if (ami_timerdev == 0) {
            // all ok
            break;
          }
        }
      }
      exit(-1);

    } while (0);
  }

  ami_tr->tr_node.io_Message.mn_ReplyPort = &replyport;
  ami_tr->tr_node.io_Command = TR_ADDREQUEST;
  ami_tr->tr_time = *tv;
  SetSignal(0, SIGF_SINGLE);
  SendIO(&ami_tr->tr_node);

  sigs = Wait(SIGBREAKF_CTRL_C | SIGF_SINGLE);
  if (!(sigs & SIGF_SINGLE)) {
    AbortIO(&ami_tr->tr_node);
  }
  WaitIO(&ami_tr->tr_node);

  if (sigs & SIGBREAKF_CTRL_C) {
    if (sig2handler) {
      sig2handler(0);
    }
  }
}

void uade_sleep(int seconds) {
  struct timeval tv = {seconds, 0};
  ami_sleep(UNIT_VBLANK, &tv);
}


void uade_usleep(int microseconds) {
  struct timeval tv = {microseconds / 1000000, microseconds % 1000000};
  ami_sleep(UNIT_MICROHZ, &tv);
}


/* returns number of milliseconds from the last call
   exception: on the first time this returns zero    */
int uade_timer_poll(void) {
  errno = ENOSYS;
  return 0;
}

char *uade_version(void) {
  return strdup(UADEVERSIONSTRING);
}

void uade_write_to_outpipe(void *ptr, int size) {
  if(unix_usingoutpipe) {
    write(unix_soundfd, ptr, size);
  }
}

